/*
 *	SynCom -- Pascal Lexer
 *
 *	(c) 1999 Martin Mares <mj@ucw.cz>
 *
 *	Can be freely distributed and used under the terms of the GNU GPL.
 */

%{
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "syncom.h"
#include "parse.tab.h"

#define YY_INPUT(buf,result,max) result = read_input(buf, max);
#define YY_NO_UNPUT
#define YY_FATAL_ERROR(msg) error(msg)

static int cmt_nest;
%}

%option noyywrap

%x COMMENT COMM2

ALPHA [a-zA-Z_]
DIGIT [0-9]
XDIGIT [0-9a-fA-F]
ALNUM [a-zA-Z_0-9]
WHITE [ \t]

%%

:=	return T_ASSIGN;
\<>	return T_NE;
\<=	return T_LE;
\>=	return T_GE;
\.\.	return T_DBLDOT;

{DIGIT}+/\.\. {
  /* This is an ugly hack to circumnvent Pascal lexical ambiguity in '1..3' */
  yylval = new_leaf(intern(yytext, L_NUM));
  return L_NUM;
}

({DIGIT}+|\${XDIGIT}+|{DIGIT}+\.{DIGIT}*|\.{DIGIT}+)([eE][+-]?{DIGIT}+)? {
  yylval = new_leaf(intern(yytext, L_NUM));
  return L_NUM;
}

{ALPHA}{ALNUM}* {
  struct keyword *w = intern(yytext, 0);
  if (w->id)		/* It's a token */
    return w->id;
  yylval = new_leaf(w);
  return L_SYM;
}

'(''|[^'\n])*' {
  yylval = new_leaf(intern(yytext, L_STRING));
  return L_STRING;
}

\{	BEGIN(COMMENT);
\(\*	cmt_nest++; BEGIN(COMM2);

[@*()+=\[\]:;<>,./^#-] {
  return yytext[0];
}

'	error("Unterminated string");

<INITIAL><<EOF>>	return X_EOF;

{WHITE}+

\n	lino++;

.	error("Unknown character");

<COMMENT>\}	BEGIN(INITIAL);
<COMMENT>\n	lino++;
<COMMENT>\{	error("Comment nesting not supported");
<COMMENT><<EOF>> error("Unterminated comment");
<COMMENT>.

<COMM2>\*\)	if (!--cmt_nest) BEGIN(INITIAL);
<COMM2>\n	lino++;
<COMM2>\(\*	cmt_nest++;
<COMM2><<EOF>>	error("Unterminated comment");
<COMM2>.

%%
