
(*
  cbasic.pas, c-style basic compiler/interpreter.
  version 1.00a, 28.6.1999
  Miroslav Spousta <spousta@bigfoot.com>
*)

{$A+,B-,D+,E+,F+,G-,I+,L+,N-,O-,R+,S+,V+,X-}
{$M 16384,0,655360}

program basic(input, output);

                        (***** Global Constants *****)

const
  cb_version = '1.00a';
  cb_ver_fsign = 'B10';                 {length(cb_ver_fsign)=3!}
  cb_date = '28.6.1999 17:07:23 CET';

  dbg = 0;
  case_sensitive = false;
  compiled_size = 16384;                {initial size in bytes}
  variable_size = 16384;
  plus_size = 16384;                    {when overflow increase size}
  stack_size = 100;                     {fields}
  stack_plus_size = 100;                {# of fields to add if overflow}
  hash_tab_size = 1000;                 {fields}
  fbuffer_size = 16384;                 {size of buffer for src file}
  label_size_max = 32;                  {max. label size}
  string_size_max = 255;                {max. string size}

  pi = 3.1415926535897932385;
  hash_const = 0.618033988;             {constant for hashing}
  cmdid   = '@';                        {cbasic command}
  funcid  = '*';                        {function}
  vfuncid = 'v';                        {variable-ret function}
  operid  = '^';                        {operator (*, /, etc.)}
  varid   = '#';                        {variable}
  clabid  = '&';                        {program label}
  dlabid  = '+';                        {data label}

  NULL = nil;
  EOFILE = chr($00);
  TABUL = chr($09);
  CR = chr($0D);
  LF = chr($0A);
  ESC = chr($1B);

  {maxint = maxint;}
  maxlong = maxlongint;
  nullstr: string[1] = '';

                       (***** Global Types *****)

type
  int = integer;                        {pascal type redefinition}
  long = longint;
  short = byte;
  float = real; {single;}
  double = real;
  strmax = string[string_size_max];
  strlab = string[label_size_max];

  p_string = ^string;                   {often used types}
  pp_string = ^p_string;
  p_char = ^char;
  p_int  = ^int;
  p_long = ^long;
  p_proc = ^t_proc;
  t_proc = procedure;

  errtype_c = (                         {basic error types}
    comp, inter, main
  );

  error_c = (                           {compile error codes}
    m_param,
    m_lowmem,
    m_opensrc,
    m_opendst,
    m_readsrc,
    m_writedst,
    m_wfile,
    c_lowmem,
    c_nocmd,
    c_expr,
    c_func,
    c_lpar,
    c_rpar,
    c_semic,
    c_comma,
    c_arrex,
    c_nvarex,
    c_nexpex,
    c_while,
    c_array,
    c_mism,
    c_redef,
    c_assign,
    c_param,
    c_type,
    c_2lab,
    c_else,
    c_label,
    i_lowmem,                           {interpreter error codes}
    i_invucmd,
    i_stacko,
    i_end,
    i_stacku,
    i_stop,
    i_smism,
    i_range,
    i_number,
    i_param,
    i_divz,
    i_conv,
    i_dims,
    i_uarray,
    i_wconst,
    i_numf,
    i_read,
    i_data,
    i_char,
    i_index
  );

  ucmd_c = (                            {u-command codes}
    jump, jmpt, jmpf, call, ret,
    move, newa, readv, restore,
    input, output,
    at, tab, cls, randomi, stop, okend
  );

  ufunop_c = (
    usin, ucos, utan, uasn, uacs, uatn,
    usqr, uln, uexp, usgn, uint, uabs, urnd,
    uchrs, uidxs, uvidxs, ulen,
    uuminus, uneg, utimes, udiv, umod, uplus, uminus,
    uloeq, uhieq, ulo, uhi, ueq, uneq, uand, uor,
    usplus, usloeq, ushieq, uslo, ushi, useq, usneq
  );

  p_ucmd = ^t_ucmd;                     {u-command record}
  t_ucmd = record
    line: long;                         {line number}
    col:  short;
    ucmd: ucmd_c;			{u-command}
  end;

  p_array = ^t_array;
  t_array = record                      {array record in 'vars' table}
    ptr: pointer;
    dims: short;
  end;

  p_number = ^t_number;
  t_number = record                     {number (int/float)}
      l: long;
      f: float;
  end;

  p_uexpr = ^t_uexpr;                   {u-expression}
  t_uexpr = record
    flags: short;
      funcop:  ufunop_c;           {infdex in u-function table}
      retadr:  long;               {uprogram+retadr = ^ucmd}
      n_array: long;               {uvars+n_array = ^t_array}
      s_array: long;               {uvars+s_array = ^t_array}
      n_const: t_number;           {number}
      s_const: long;               {uvars+s_const = ^string}
      n_var:   long;               {uvars+n_var = ^t_uexpr}
      s_var:   long;               {uvars+s_var = ^^string}
      n_eval:  t_number;           {evaluated number}
      s_eval:  p_string;           {evaluated pointer at string constant}
      n_varp:  p_uexpr;            {^p_uexpr}
      s_varp:  pp_string;          {^^string}
      c_varp:  p_char;             {^char}
  end;

const
  {masks for type expr}
  ue_retadr = $80;                      {return address/not}
  ue_last   = $40;                      {last expr/not}
  ue_funcop = $20;                      {function/not}
  ue_array  = $10;                      {array/single variable}
  ue_evaled = $08;                      {ready to process/not}
  ue_const  = $04;                      {constant/variable}
  ue_string = $02;                      {string/number}
  ue_float  = $01;                      {float/long}
  ue_temp   = $01;                      {temp-$/not}
  ue_char   = $01;                      {only one character/not}

  {masks for global flags}
  gf_conly  = $01;
  gf_eonly  = $02;


(* Compiler Type Definitions *)

type
  p_fproc = ^t_fproc;
  t_fproc = procedure;

  hflag_c = (
    done, forw
  );

  p_hforw = ^t_hforw;                   {hash -- forward chain entry}
  t_hforw = record
    next: p_hforw;
    ofs:  long;
  end;

  p_hlab = ^t_hlab;
  t_hlab = record
    flag: hflag_c;
    ofs: long;
    forw: p_hforw;
  end;

  p_hash = ^t_hash;                     {hash table}
  t_hash = record
    next: p_hash;
    name: string[label_size_max];     {32+1 for recognization flag}
      cmd: t_proc;                 {command}
      func: t_fproc;               {function}
      oper: t_fproc;
         pcmp: short; psto: short;    {operator}
      clab: t_hla;                {command label}
      dlab: t_hlab;                {data label}
      vari: t_uexpr;               {variable/array}
  end;

  p_cforw = ^t_cforw;                   {break/continue forward chain entry}
  t_cforw = record
    next: p_cforw;
    level: long;
    ofs: long;
  end;

(* Global Variable Definitions *)

var
  hash_tab: array [0..hash_tab_size]  {hash table}
    of p_hash;

  uprogram: pointer;                    {CI `compiled' program}
  uvars: pointer;                       {CI variables}
  ustack: p_uexpr;                      { I uprogram stack}

  uprogram_start: long;                 {CI}
  udata_start: long;                    {CI}

  uprogram_top: long;                   {C }
  uvars_top: long;                      {C }
  udata_top: long;                      {C }

  uprogramp: long;                      { I}
  udatap: long;                         { I}
  ustackp: long;                        { I}

  chact: char;                          {C  reading from file}
  chnext: char;                         {C }

  line: long;                           {CI for error messages}
  col: short;                           {CI}

  fbuffer: pointer;                     {C }
  fbuffer_top: long;                    {C }
  fbuffer_max: long;                    {C }

  break_forwp: p_cforw;
  cont_forwp: p_cforw;

  ifdistance: long;
  cyclelevel: long;
  upgmend: boolean;

  srcfile: file;
  dstfile: file;

  gflags: short;                        {global flags}
  s: string;
  f3: string[3];
  i: int;

                         (***** Forwards *****)

procedure main_cmd; forward;            {cbasic commands}
procedure label_cmd; forward;
procedure goto_cmd; forward;
procedure let_cmd; forward;
procedure dim_cmd; forward;
procedure print_cmd; forward;
procedure input_cmd; forward;
procedure data_cmd; forward;
procedure read_cmd; forward;
procedure restore_cmd; forward;
procedure if_cmd; forward;
procedure else_cmd; forward;
procedure for_cmd; forward;
procedure while_cmd; forward;
procedure do_cmd; forward;
procedure break_cmd; forward;
procedure continue_cmd; forward;
procedure sub_cmd; forward;
procedure return_cmd; forward;
procedure gosub_cmd; forward;
procedure at_cmd; forward;
procedure tab_cmd; forward;
procedure cls_cmd; forward;
procedure randomize_cmd; forward;
procedure stop_cmd; forward;
                                        {cbasic functions}
procedure sin_fn(var valifo: pointer); forward;
procedure cos_fn(var valifo: pointer); forward;
procedure tan_fn(var valifo: pointer); forward;
procedure asn_fn(var valifo: pointer); forward;
procedure acs_fn(var valifo: pointer); forward;
procedure atn_fn(var valifo: pointer); forward;
procedure sqr_fn(var valifo: pointer); forward;
procedure ln_fn(var valifo: pointer); forward;
procedure exp_fn(var valifo: pointer); forward;
procedure sgn_fn(var valifo: pointer); forward;
procedure int_fn(var valifo: pointer); forward;
procedure abs_fn(var valifo: pointer); forward;
procedure rnd_fn(var valifo: pointer); forward;
procedure chrs_fn(var valifo: pointer); forward;
procedure idxs_fn(var valifo: pointer); forward;
procedure vidxs_fn(var valifo: pointer); forward;
procedure len_fn(var valifo: pointer); forward;
                                        {cbasic operators}
procedure uplus_op(var valifo: pointer); forward;
procedure uminus_op(var valifo: pointer); forward;
procedure neg_op(var valifo: pointer); forward;
procedure times_op(var valifo: pointer); forward;
procedure div_op(var valifo: pointer); forward;
procedure mod_op(var valifo: pointer); forward;
procedure plus_op(var valifo: pointer); forward;
procedure minus_op(var valifo: pointer); forward;
procedure loeq_op(var valifo: pointer); forward;
procedure hieq_op(var valifo: pointer); forward;
procedure lo_op(var valifo: pointer); forward;
procedure hi_op(var valifo: pointer); forward;
procedure eq_op(var valifo: pointer); forward;
procedure neq_op(var valifo: pointer); forward;
procedure and_op(var valifo: pointer); forward;
procedure or_op(var valifo: pointer); forward;

procedure jump_ucmd; forward;           {u-commands}
procedure jmpt_ucmd; forward;
procedure jmpf_ucmd; forward;
procedure call_ucmd; forward;
procedure ret_ucmd; forward;
procedure move_ucmd; forward;
procedure newa_ucmd; forward;
procedure read_ucmd; forward;
procedure restore_ucmd; forward;
procedure input_ucmd; forward;
procedure output_ucmd; forward;
procedure at_ucmd; forward;
procedure tab_ucmd; forward;
procedure cls_ucmd; forward;
procedure random_ucmd; forward;
procedure stop_ucmd; forward;
procedure end_ucmd; forward;

procedure sin_ufn; forward;             {u-functions}
procedure cos_ufn; forward;
procedure tan_ufn; forward;
procedure asn_ufn; forward;
procedure acs_ufn; forward;
procedure atn_ufn; forward;
procedure sqr_ufn; forward;
procedure ln_ufn; forward;
procedure exp_ufn; forward;
procedure sgn_ufn; forward;
procedure int_ufn; forward;
procedure abs_ufn; forward;
procedure rnd_ufn; forward;
procedure chrs_ufn; forward;
procedure idxs_ufn; forward;
procedure vidxs_ufn; forward;
procedure len_ufn; forward;

procedure uminus_uop; forward;          {u-operators}
procedure neg_uop; forward;
procedure times_uop; forward;
procedure div_uop; forward;
procedure mod_uop; forward;
procedure plus_uop; forward;
procedure minus_uop; forward;
procedure loeq_uop; forward;
procedure hieq_uop; forward;
procedure lo_uop; forward;
procedure hi_uop; forward;
procedure eq_uop; forward;
procedure neq_uop; forward;
procedure and_uop; forward;
procedure or_uop; forward;
procedure splus_uop; forward;
procedure sloeq_uop; forward;
procedure shieq_uop; forward;
procedure slo_uop; forward;
procedure shi_uop; forward;
procedure seq_uop; forward;
procedure sneq_uop; forward;

     (******************** Global Functions ********************)

                        (***** Debug *****)

procedure debug(s: strmax);
begin
  if (dbg <> 0) then
    writeln('DBG: ', s);
end;

                   (***** Error Handling *****)

procedure error(e: errtype_c; c: error_c; line: long; col: short);
var
  xs: string[50];
begin
  writeln;
  write('cbasic ');
  case e of
    comp:  write('compiler');
    inter: write('interpreter');
    main:  write('error');
  end;
  xs:= 'Unknown error';
  case c of
    m_param:    xs:= 'Invalid parameter';
    m_lowmem:   xs:= 'Not enough memory';
    m_opensrc:  xs:= 'Cannot open input file';
    m_opendst:  xs:= 'Cannot open output file';
    m_readsrc:  xs:= 'Error reading input file';
    m_writedst: xs:= 'Error writing ouptut file';
    m_wfile:    xs:= 'Unknown file format';
    c_lowmem:   xs:= 'Not enough memory';
    c_nocmd:    xs:= 'Unknown command';
    c_expr:     xs:= 'Invalid expression';
    c_func:     xs:= 'Unknown function';
    c_lpar:     xs:= '"(" expected';
    c_rpar:     xs:= '")" expected';
    c_semic:    xs:= '";" expected';
    c_comma:    xs:= '"," expected';
    c_assign:   xs:= '"=" expected';
    c_arrex:    xs:= 'Array expected';
    c_nvarex:   xs:= 'Numeric variable expected';
    c_nexpex:   xs:= 'Numeric expression expected';
    c_while:    xs:= 'WHILE expected';
    c_array:    xs:= 'Invalid array index';
    c_mism:     xs:= 'Illegal redefinition';
    c_redef:    xs:= 'Already defined elsewhere';
    c_param:    xs:= 'Invalid parameter';
    c_type:     xs:= 'Invalid operand';
    c_2lab:     xs:= 'Duplicate label';
    c_else:     xs:= 'unexpected ELSE';
    c_label:    xs:= 'Undefined forward';
    i_lowmem:   xs:= 'Not enough memory';
    i_stop:     xs:= 'Program stop';
    i_end:      xs:= 'Program end';
    i_stacko:   xs:= 'Stack overflow';
    i_stacku:   xs:= 'Stack underflow';
    i_invucmd:  xs:= 'Invalid u-command';
    i_smism:    xs:= 'Stack mismatch';
    i_range:    xs:= 'Integer too big';
    i_number:   xs:= 'Number expected';
    i_param:    xs:= 'Invalid parameter encountered';
    i_divz:     xs:= 'Result not defined';
    i_conv:     xs:= 'Cannot convert to integer';
    i_dims:     xs:= 'Invalid array usage';
    i_uarray:   xs:= 'Array not initialized';
    i_wconst:   xs:= 'Variable expected (WC)';
    i_numf:     xs:= 'Wrong number';
    i_read:     xs:= 'Out of data';
    i_data:     xs:= 'Incompatible data';
    i_char:     xs:= 'Cannot convert to char';
    i_index:    xs:= 'Invalid index';
  end;
  write(': ', xs);
  if (line <> 0) then
    writeln(' [', line, ':', col, ']')
  else
    writeln('!');
  halt(ord(c));
end;

                  (***** Character Functions *****)

function isalpha(c: char): boolean;
begin
  isalpha:= (((c >= 'A') and (c <= 'Z')) or
    ((c >= 'a') and (c <= 'z')) or (c = '_'));
end;

function isnum(c: char): boolean;
begin
  isnum:= ((c >= '0') and (c <= '9'));
end;

function isalnum(c: char): boolean;
begin
  isalnum:= (isnum(c) or isalpha(c));
end;

function isoper(c: char): boolean;
begin
  isoper:= ((c = '!') or (c = '|') or ((c <> ',') and
          (((c > '"') and (c < '(')) or
           ((c > ')') and (c < '0')) or
           ((c > ';') and (c < 'A')))));
end;

function upstr(s: strmax): strmax;
var
  i: int;
begin
  for i:= 1 to length(s) do
    s[i]:= upcase(s[i]);
  upstr:= s;
end;

                   (***** Memory Functions *****)

function malloc(size: long): pointer;
var
  xp: ^long;
  xl: long;
begin
  xl:= size + sizeof(long);
  if (xl <= maxavail) then begin
    getmem(xp, xl);
    xp^:= xl;
    inc(xp)
  end else
    xp:= NULL;

  malloc:= xp;
end;

procedure mfree(var p: pointer);
var
  xl: long;
  xp: ^long;
begin
  xp:= p;
  dec(xp);
  freemem(xp, xp^);
  p:= NULL;                             {for security reasons}
end;

function realloc(var p: pointer; size: long): pointer;
var
  ps, pd: ^byte;
  px: ^long;
  il, l: long;
begin
  pd:= malloc(size);
  realloc:= pd;
  if (pd <> NULL) then begin
    px:= p;
    dec(px);
    l:= px^;
    if (l > size) then
      l:=  size;
    ps:= p;
    for il:= 1 to l do begin
      pd^:= ps^;
      inc(ps);
      inc(pd);
    end;
    mfree(p);
  end;
end;

                   (***** uexpr Functions ****)

function isint(var ue: t_uexpr): boolean;
begin
  isint:= (ue.flags and ue_float) = 0;
end;

function both_ints(var ue1, ue2: t_uexpr): boolean;
begin
  both_ints:= (((ue1.flags and ue_float) = 0)
      and ((ue2.flags and ue_float) = 0));
end;

procedure int2float(var ue: t_uexpr);
begin
  if ((ue.flags and ue_float) = 0) then begin
    ue.n_eval.f:= ue.n_eval.l;
    ue.flags:= ue.flags or ue_float;
  end;
end;

procedure try2int(var ue: t_uexpr);
var
  l: long;
  f: float;
begin
  if ((ue.flags and ue_float) <> 0) then begin
    f:= ue.n_eval.f;
    if ((f < maxlong) and (f > maxlong)) then begin
      l:= round(f);
      if (l = f) then begin
        ue.n_eval.l:= l;
        ue.flags:= ue.flags xor ue_float;
      end;
    end;
  end;
end;

function do2int(var ue: t_uexpr): boolean;
var
  xf: float;
begin
  do2int:= true;
  if ((ue.flags and ue_float) <> 0) then begin
    xf:= ue.n_eval.f;
    if ((xf < maxlong) and (xf > maxlong)) then begin
      ue.flags:= ue.flags xor ue_float;
      ue.n_eval.l:= trunc(xf);
    end else
      do2int:= false;
  end;
end;

function iszero(var ue: t_uexpr): boolean;
begin
  iszero:= ((((ue.flags and ue_float) <> 0) and (ue.n_eval.f = 0)) or
     (((ue.flags and ue_float) = 0) and (ue.n_eval.l = 0)));
end;


      (********************  COMPILER ********************)

type
  str_c = (                             {gets type}
    str, lab
  );

  nums_c = (                            {getn}
    nint, nfrac, nexp
  );

  expect_c = (                          {process_expr}
    value, oper
  );

  par_c = (                             {parenthesis}
    left, right
  );

  val_c = (                             {process_expr return code}
    vempty, vnum, vstr
  );

  process_c = (                         {what process}
    variable, expression
  );

  p_valifo = ^t_valifo;                 {last value type chain}
  t_valifo = record
    next: p_valifo;
    what: val_c;
  end;

  p_oplifo = ^t_oplifo;                 {operators stack}
  t_oplifo = record
    next: p_oplifo;
    oper: t_fproc;
    psto: short;                        {stored-priority}
  end;

  t_cmd = record                        {init hash table}
    name: string[9];
    cmd: t_proc;
  end;

  t_func = record                       {init hash table}
    name: string[6];
    func: t_fproc;
  end;

  t_oper = record                       {init hash table}
    name: string[2];
    oper: t_fproc;
    pcmp: short;
    psto: short;
  end;

const
  hash_cmd_size = 25;                   {init hash table}


            (***** Memory Procedure ****)

function cnew(len: long): pointer;      {this is compiler procedure}
var
  px: pointer;
begin
  px:= malloc(len);
  if (px = NULL) then
    error(comp, c_lowmem, line, col);
  cnew:= px;
end;


              (***** Managing Hash Table *****)

function hash(s: strlab): long;         {main hash function}
var
  i, sum: int;
begin
  sum:= 0;
  for i:= 1 to length(s) do
    sum:= sum + ord(s[i]);
  hash:= trunc(hash_tab_size*frac(hash_const*sum));
end;                                    {return 0..hash_tab_size-1}

function srch_hash(s: strlab): p_hash;
var
  px: p_hash;
  found: boolean;
begin
  if (not case_sensitive) then
    s:= upstr(s);
  px:= hash_tab[hash(s)];
  while ((px <> NULL) and (s <> px^.name)) do
    px:= px^.next;
  srch_hash:= px;
end;

procedure add2hash(h: p_hash);
var
  index: long;
begin
  if (not case_sensitive) then
    h^.name:= upstr(h^.name);
  index:= hash(h^.name);
  h^.next:= hash_tab[index];
  hash_tab[index]:= h;
end;

procedure init_hash;                    {fill hash table with cmds and funcs}
var
  li: long;
  i: int;
  hx: p_hash;
  fx: t_func;
  ox: t_oper;
begin
  for li:= 0 to hash_tab_size-1 do      {begin -- set everything to NULL}
    hash_tab[li]:= NULL;
  for i:= 1 to hash_cmd_size do begin
    hx:= cnew(sizeof(hx^));
    hx^.name:= cmdid + hash_cmd_tab[i].name;
    hx^.cmd:= hash_cmd_tab[i].cmd;
    add2hash(hx);
  end;
  for i:= 1 to hash_func_size do begin
    hx:= cnew(sizeof(hx^));
    fx:= hash_func_tab[i];
    hx^.name:= funcid + fx.name;
    hx^.func:= fx.func;
    add2hash(hx);
  end;
  for i:= 1 to hash_vfunc_size do begin
    hx:= cnew(sizeof(hx^));
    fx:= hash_vfunc_tab[i];
    hx^.name:= vfuncid + fx.name;
    hx^.func:= fx.func;
    add2hash(hx);
  end;
  for i:= 1 to hash_oper_size do begin
    hx:= cnew(sizeof(hx^));
    ox:= hash_oper_tab[i];
    hx^.name:= operid + ox.name;
    hx^.oper:= ox.oper;
    hx^.pcmp:= ox.pcmp;
    hx^.psto:= ox.psto;
    add2hash(hx);
  end;
end;

procedure free_hash;                    {free memory used by hash}
var
  li: int;
  ph, phx: p_hash;
  ch: char;
begin
  for li:= 0 to hash_tab_size-1 do begin
    ph:= hash_tab[li];
    while (ph <> NULL) do begin
      ch:= ph^.name[1];
      if (ch = clabid) then begin
        if (ph^.clab.flag = forw) then begin
          writeln;
          write('label/sub: ', copy(ph^.name, 2, 255));
          error(comp, c_label, 0, 0);
        end;
      end
      else if(ch = dlabid) then begin
        if (ph^.dlab.flag = forw) then begin
          writeln;
          write('data label: ', copy(ph^.name, 2, 255));
          error(comp, c_label, 0, 0);
        end;
      end;
      phx:= ph^.next;
      mfree(pointer(ph));
      ph:= phx;
    end;
  end;
end;

                   (***** Buffered Read *****)

function get_one: char;
var
  result: int;
  px: ^char;
begin
  if (fbuffer_top = fbuffer_max) then begin
    blockread(srcfile, fbuffer^, fbuffer_size, result);
    fbuffer_top:= 0;
    fbuffer_max:= result;
  end;
  if (fbuffer_top = fbuffer_max) then
    get_one:= EOFILE
  else begin
    px:= pointer(fbuffer);
    inc(px, fbuffer_top);
    get_one:= px^;
    inc(fbuffer_top);
  end;
end;

procedure nextch;                       {bypass one char}
begin
  inc(col);
  if (chact = LF) then begin
    inc(line);
    col:= 1;
  end;
  chact:= chnext;
  chnext:= get_one;
end;

function lookch: char;                  {gets one char; next time this same}
begin
  lookch:= chact;
end;

function getch: char;                   {gets one char; next time next one}
begin
  getch:= chact;
  nextch;
end;

procedure eat_ws;                       {ignore white spaces and comments}
var
  nows, nocom: boolean;
begin
  nocom:= true;
  repeat
    nows:= true;
    if ((chact = '/') and (chnext = '*')) then begin
      nocom:= false;
      nextch;
    end;
    if ((chact = '*') and (chnext = '/')) then begin
      nocom:= true;
      nextch; nextch;
    end;
    if (nocom = false) then
      nextch;
    if ((chact = ' ') or (chact = TABUL) or
       (chact = CR) or (chact = LF)) then begin
      nextch;
      nows:= false;
    end;
  until ((nows and nocom) or (chact = EOFILE));
end;

function gets(t: str_c): strmax;        {get string (label, ...)}
var
  terminate: boolean;
  len: int;
  ch: char;
  s: strmax;
begin
  s:= '';
  if (t = str) then begin
    len:= string_size_max;
    nextch;                             {ignore begining '"'}
  end else begin
    len:= label_size_max;
    eat_ws;
  end;
  terminate:= false;
  repeat
    ch:= lookch;
    if (t = str) then begin             {string}
      if (ch <> '"') then
        s:= s + ch
      else
        terminate:= true;
      nextch;
    end else begin                      {label}
      if (isalnum(ch)) then begin
        s:= s + ch;
        nextch;
      end else
        terminate:= true;
    end;
    dec(len);
  until ((terminate) or (len = 0));
  gets:= s;
end;

procedure getn(var ue: t_uexpr);
var
  ch: char;
  terminate: boolean;
  status: nums_c;
  numpart, xfrac: double;
  exppart: int;
  xl: long;
  v: short;
  i: int;
begin
  numpart:= 0;
  exppart:= 0;
  terminate:= false;
  status:= nint;
  repeat
    ch:= lookch;
    if (ch = '.') then begin
      status:= nfrac;
      xfrac:= 0.1;
      nextch;
    end else if (upcase(ch) = 'E') then begin
      status:= nexp;
      nextch;
    end else if (isnum(ch)) then begin
      v:= ord(ch) - ord('0');
      nextch;
      if (status = nint) then           {int part}
        numpart:= numpart * 10 + v
      else if (status = nfrac) then begin
        numpart:= numpart + xfrac * v;
        xfrac:= xfrac / 10;             {frac part}
      end else
        exppart:= exppart * 10 + v;
    end else
      terminate:= true;
  until (terminate);

  for i:= 1 to exppart do begin
    numpart:= numpart * 10;
  end;
  ue.flags:= ue_const or ue_float;
  ue.n_const.f:= numpart;
  if (status <> nfrac) then
    try2int(ue);
end;

                    (***** Putting Structures *****)

procedure puto(src: pointer; len: int; var dst: pointer; var ofs: long);
var
  i: int;
  px: ^long;
  psrc, pdst: ^char;
begin
  psrc:= src;
  pdst:= dst;
  inc(pdst, ofs);
  ofs:= ofs + len;
  px:= dst;
  dec(px);
  if (px^ < ofs) then begin             {expand when necessary}
    dst:= realloc(dst, px^ + plus_size);
    if (dst = NULL) then
      error(comp, c_lowmem, line, col);
  end;
  for i:= 1 to len do begin             {copy loop}
    pdst^:= psrc^;
    inc(psrc);
    inc(pdst);
  end;
end;

procedure put_upgm(o: pointer; len: int);
begin
  puto(o, len, uprogram, uprogram_top);
end;

procedure put_uvars(o: pointer; len: int);
begin
  puto(o,len, uvars, uvars_top);
end;

procedure mov_upgm(ofs: long; o: pointer; len: int);
var
  xl: long;
begin
  xl:= uprogram_top;
  uprogram_top:= ofs;
  puto(o, len, uprogram, uprogram_top);
  uprogram_top:= xl;
end;

procedure mov_uvars(ofs: long; o: pointer; len: int);
var
  xl: long;
begin
  xl:= uvars_top;
  uvars_top:= ofs;
  puto(o,len, uvars, uvars_top);
  uvars_top:= xl;
end;

               (***** Process Expression *****)

procedure add2valifo(var lifop: p_valifo; what: val_c);
var
  pv: p_valifo;
begin
  pv:= cnew(sizeof(pv^));
  pv^.next:= lifop;
  pv^.what:= what;
  lifop:= pv;
end;

function get_valifo(var lifop: p_valifo): val_c;
var
  pv: p_valifo;
begin
  if (lifop <> NULL) then begin
    pv:= lifop;
    lifop:= pv^.next;
    get_valifo:= pv^.what;
    mfree(pointer(pv));
  end else
    get_valifo:= vempty;
end;
                                        {true -- at least one done}
function process_expr(what: process_c): val_c;

var
  ch: char;
  s: strlab;
  i: int;
  expect: expect_c;
  isstr, terminate: boolean;
  oplifop: p_oplifo;
  valifop: p_valifo;
  last_top: long;
  wrong: boolean;
  parcounter: long;
  xfuncid: string[1];
  ue: t_uexpr;
  ph: p_hash;
  ps: ^string;
  pc: ^char;
  po: p_oplifo;
  ar: t_array;
  xb: boolean;

  procedure add2oplifo(var valifop: p_valifo; h: p_hash);
  var                                   {add to operators stack}
    xp, xpo: p_oplifo;
  begin
    xp:= oplifop;                       {needs short expr. eval.}
    while ((xp <> NULL) and (xp^.psto <= h^.pcmp)) do begin
      xpo:= xp;
      xp:= xp^.next;
      mfree(pointer(xpo));
    end;
    xpo:= cnew(sizeof(xpo^));
    xpo^.next:= xp;
    xpo^.oper:= h^.oper;
    xpo^.psto:= h^.psto;
    oplifop:= xpo;
  end;

begin
  oplifop:= NULL;                       {operators stack}
  valifop:= NULL;                       {last values chain}
  last_top:= 0;                         {actual last part of expression}
  parcounter:= 0;
  if (what <> variable) then begin
    terminate:= false;                  {read full expression}
    xfuncid:= funcid;
  end else begin
    terminate:= true;                   {read only var address}
    xfuncid:= vfuncid;
  end;
  expect:= value;
  repeat
    eat_ws;
    ch:= lookch;
    if (expect = value) then begin      {---------- value}
      expect:= oper;
      if (isalpha(ch)) then begin
        s:= gets(lab);
        isstr:= (lookch = '$');         {is it string var/function}
        if (isstr) then begin
          s:= s + '$';
          nextch;
        end;
        ph:= srch_hash(xfuncid + s);    {FUNCTION?}
        if (ph <> NULL) then begin
          last_top:= uprogram_top + sizeof(t_uexpr);
        end else begin
          ph:= srch_hash(varid + s);    {VARIABLE?}
          eat_ws;
          if (lookch <> '[') then begin
            if (ph <> NULL) then begin
              if ((ph^.vari.flags and ue_array) <> 0) then
                error(comp, c_mism, line, col);
            end                         {already known var}
            else begin
              ph:= cnew(sizeof(ph^));
              ph^.name:= varid + s;
              if (not isstr) then begin
                ph^.vari.flags:= 0;     {new numeric variable}
                ph^.vari.n_var:= uvars_top;
                add2hash(ph);
                ue.flags:= ue_const;
                ue.n_const.l:= 0;
                put_uvars(@ue, sizeof(ue));
              end
              else begin                {new string variable}
                ph^.vari.flags:= ue_string;
                ph^.vari.s_var:= uvars_top;
                add2hash(ph);
                ps:= NULL;
                put_uvars(@ps, sizeof(ps));
              end;
            end;                        {store pointer to variable}
            put_upgm(@ph^.vari, sizeof(ph^.vari));
            last_top:= uprogram_top;
          end
          else begin                    {ARRAY}
            nextch;
            if (ph <> NULL) then begin
              if ((ph^.vari.flags and ue_array) = 0) then
                error(comp, c_mism, line, col);
            end                         {already known array}
            else begin
              ph:= cnew(sizeof(ph^));
              ph^.name:= varid + s;     {new numeric array}
              if (not isstr) then begin
                ph^.vari.flags:= ue_array;
                ph^.vari.n_array:= uvars_top;
              end
              else begin                {new string array}
                ph^.vari.flags:= ue_array or ue_string;
                ph^.vari.s_array:= uvars_top;
              end;
              add2hash(ph);
              ar.ptr:= NULL;
              ar.dims:= 0;              {important! 0 = uninitialized}
              put_uvars(@ar, sizeof(ar));
            end;
            put_upgm(@ph^.vari, sizeof(ph^.vari));
            last_top:= uprogram_top;    {last is array}

            put_upgm(@i, sizeof(i));    {space for number of dimensions (i)}
            i:= 0;
            repeat
              wrong:= (process_expr(expression) <> vnum);
              inc(i);
              eat_ws;
              ch:= getch;
            until (wrong or (ch <> ','));
            if (wrong or (ch <> ']')) then
              error(comp, c_array, line, col);
            mov_upgm(last_top, @i, sizeof(i)); {# of array dims}
          end;
          if (isstr) then               {for both vars and arrays}
            add2valifo(valifop, vstr)   {last value depends on isstr}
          else
            add2valifo(valifop, vnum);
        end;
      end
      else if (what <> variable) then begin
        if (ch = '(') then begin        {LEFT PARENTHESIS}
          expect:= value;
          nextch;
          po:= cnew(sizeof(t_oplifo));
          po^.next:= oplifop;
          po^.psto:= 100;               {will not be kicked out}
          oplifop:= po;                 {func ... doesn't bother me}
          inc(parcounter);
        end
        else if ((ch = '!') or (ch = '-') or (ch = '+')) then begin
          expect:= value;
          nextch;
          ph:= srch_hash(operid + ch + '~');
          if (ph = NULL) then           {! or plus or minus before number}
            error(comp, c_func, line, col);
          add2oplifo(valifop, ph);
          last_top:= uprogram_top;
        end
        else if (isnum(ch) or (ch = '.')) then begin
          getn(ue);                     {NUMERIC CONSTANT}
          put_upgm(@ue, sizeof(ue));
          last_top:= uprogram_top;
          add2valifo(valifop, vnum);
        end
        else if (ch = '"') then begin
          s:= gets(str);                {STRING CONSTANT}
          ue.flags:= (ue_const or ue_string);
          ue.s_const:= uvars_top;
          put_upgm(@ue, sizeof(ue));
          put_uvars(@s, length(s)+1);
          last_top:= uprogram_top;
          add2valifo(valifop, vstr);
        end else                        {anything else...}
          error (comp, c_expr, line, col);
      end;
    end
    else begin                          {---------- operator}
      if (ch = ')') then begin          {RIGHT PARENTHESIS}
        if (parcounter > 0) then begin
          nextch;
          po:= oplifop;                 {cannot reach NULL}
          while (po^.psto < 100) do begin
            last_top:= uprogram_top;
            p_oplifo(pc):= po;
            po:= po^.next;
            mfree(pointer(pc));
          end;
          oplifop:= po^.next;
          mfree(pointer(po));
        end else
          terminate:= true;             {not my left par.}
      end
      else if (isoper(ch)) then begin
        expect:= value;                 {OPERATOR}
        s:= operid;
        repeat
          s:= s + ch;
          nextch;
          eat_ws;
          ch:= lookch;
        until ((not isoper(ch)) or (ch = '+') or (ch = '-'));
        ph:= srch_hash(s);
        if (ph = NULL) then
          error(comp, c_expr, line, col);
        add2oplifo(valifop, ph);
        last_top:= uprogram_top;
      end else
        terminate:= true;               {anything else...}
    end;
  until (terminate);

  while (oplifop <> NULL) do begin
    if (oplifop^.psto >= 100) then
      error(comp, c_expr, line, col);    {too many left parenth's}
    last_top:= uprogram_top;
    po:= oplifop;
    oplifop:= oplifop^.next;
    mfree(pointer(po));
  end;

  process_expr:= get_valifo(valifop);
  if (valifop <> NULL) then             {something stayed on stack...}
    error(comp, c_expr, line, col);
  if (last_top <> 0) then begin
    pc:= uprogram;                      {mark last t_uexpr as 'last' one}
    inc(pc, last_top - sizeof(t_uexpr));
    p_uexpr(pc)^.flags:= p_uexpr(pc)^.flags or ue_last;
  end;
end;

               (***** cbasic Functions *****)

procedure put_funcop(funcop: ufunop_c);
var
  ue: t_uexpr;
begin
  ue.flags:= ue_funcop;
  ue.funcop:= funcop;
  put_upgm(@ue, sizeof(ue));
end;

procedure check_lpar;
begin
  eat_ws;
  if (getch <> '(') then
    error(comp, c_lpar, line, col);     {no left parenthesis found}
end;

procedure check_rpar;
begin
  eat_ws;
  if (getch <> ')') then
    error(comp, c_rpar, line, col);     {no left parenthesis found}
end;

procedure check_comma;
begin
  eat_ws;
  if (getch <> ',') then
    error(comp, c_comma, line, col);
end;

procedure eval_numparam;
begin
  check_lpar;
  if (process_expr(expression) <> vnum) then
    error(comp, c_param, line, col);    {invalid parameter}
  check_rpar;
end;

procedure eval_strparam;
begin
  check_lpar;
  if (process_expr(expression) <> vstr) then
    error(comp, c_param, line, col);    {invalid parameter}
  check_rpar;
end;

procedure sin_fn(var valifo: pointer);  {SIN}
begin
  put_funcop(usin);
  eval_numparam;
  add2valifo(p_valifo(valifo), vnum);
end;

procedure cos_fn(var valifo: pointer);  {COS}
begin
  put_funcop(ucos);
  eval_numparam;
  add2valifo(p_valifo(valifo), vnum);
end;

procedure tan_fn(var valifo: pointer);  {TAN}
begin
  put_funcop(utan);
  eval_numparam;
  add2valifo(p_valifo(valifo), vnum);
end;

procedure asn_fn(var valifo: pointer);  {ASN}
begin
  put_funcop(uasn);
  eval_numparam;
  add2valifo(p_valifo(valifo), vnum);
end;

procedure acs_fn(var valifo: pointer);  {ACS}
begin
  put_funcop(uacs);
  eval_numparam;
  add2valifo(p_valifo(valifo), vnum);
end;

procedure atn_fn(var valifo: pointer);  {ATN}
begin
  put_funcop(uatn);
  eval_numparam;
  add2valifo(p_valifo(valifo), vnum);
end;

procedure sqr_fn(var valifo: pointer);  {SQR}
begin
  put_funcop(usqr);
  eval_numparam;
  add2valifo(p_valifo(valifo), vnum);
end;

procedure ln_fn(var valifo: pointer);   {LN}
begin
  put_funcop(uln);
  eval_numparam;
  add2valifo(p_valifo(valifo), vnum);
end;

procedure exp_fn(var valifo: pointer);  {EXP}
begin
  put_funcop(uexp);
  eval_numparam;
  add2valifo(p_valifo(valifo), vnum);
end;

procedure sgn_fn(var valifo: pointer);  {SGN}
begin
  put_funcop(usgn);
  eval_numparam;
  add2valifo(p_valifo(valifo), vnum);
end;

procedure int_fn(var valifo: pointer);  {INT}
begin
  put_funcop(uint);
  eval_numparam;
  add2valifo(p_valifo(valifo), vnum);
end;

procedure abs_fn(var valifo: pointer);  {ABS}
begin
  put_funcop(uabs);
  eval_numparam;
  add2valifo(p_valifo(valifo), vnum);
end;

procedure rnd_fn(var valifo: pointer);  {RND}
begin
  put_funcop(urnd);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure chrs_fn(var valifo: pointer); {CHR$}
begin
  put_funcop(uchrs);
  eval_numparam;
  add2valifo(p_valifo(valifo), vstr);
end;

procedure idxs_fn(var valifo: pointer); {IDX$}
begin
  put_funcop(uidxs);
  check_lpar;
  if (process_expr(expression) <> vstr) then  {string}
    error(comp, c_param, line, col);
  check_comma;
  if (process_expr(expression) <> vnum) then  {number}
    error(comp, c_param, line, col);
  check_rpar;
  add2valifo(p_valifo(valifo), vstr);
end;

procedure vidxs_fn(var valifo: pointer);
begin                                   {Variable-return IDX$}
  put_funcop(uvidxs);
  check_lpar;
  if (process_expr(variable) <> vstr) then
    error(comp, c_param, line, col);    {string}
  check_comma;
  if (process_expr(expression) <> vnum) then
    error(comp, c_param, line, col);    {number}
  check_rpar;
  add2valifo(p_valifo(valifo), vstr);
end;

procedure len_fn(var valifo: pointer);  {LEN}
begin
  put_funcop(ulen);
  eval_strparam;
  add2valifo(p_valifo(valifo), vnum);
end;

procedure uplus_op(var valifo: pointer);
var                                     {+X}
  last: val_c;
begin
  last:= get_valifo(p_valifo(valifo));
  if (last <> vnum) then
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure uminus_op(var valifo: pointer);
var                                      {-X}
  last: val_c;
begin
  put_funcop(uuminus);
  last:= get_valifo(p_valifo(valifo));
  if (last <> vnum) then
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure neg_op(var valifo: pointer);
var                                     {NEG}
  last: val_c;
begin
  put_funcop(uneg);
  last:= get_valifo(p_valifo(valifo));
  if (last <> vnum) then
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure times_op(var valifo: pointer);
var                                     {TIMES}
  last1, last2: val_c;
begin
  put_funcop(utimes);
  last1:= get_valifo(p_valifo(valifo));
  last2:= get_valifo(p_valifo(valifo));
  if ((last1 <> vnum) or (last2 <> vnum)) then
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure div_op(var valifo: pointer);  {DIV}
var
  last1, last2: val_c;
begin
  put_funcop(udiv);
  last1:= get_valifo(p_valifo(valifo));
  last2:= get_valifo(p_valifo(valifo));
  if ((last1 <> vnum) or (last2 <> vnum)) then
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure mod_op(var valifo: pointer);  {MOD}
var
  last1, last2: val_c;
begin
  put_funcop(umod);
  last1:= get_valifo(p_valifo(valifo));
  last2:= get_valifo(p_valifo(valifo));
  if ((last1 <> vnum) or (last2 <> vnum)) then
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure plus_op(var valifo: pointer); {PLUS}
var
  last1, last2: val_c;
begin
  last1:= get_valifo(p_valifo(valifo));
  last2:= get_valifo(p_valifo(valifo));
  if ((last1 = vnum) and (last2 = vnum)) then begin
    put_funcop(uplus);
    add2valifo(p_valifo(valifo), vnum);
  end
  else if ((last1 = vstr) and (last2 = vstr)) then begin
    put_funcop(usplus);
    add2valifo(p_valifo(valifo), vstr);
  end else
    error(comp, c_type, line, col);
end;

procedure minus_op(var valifo: pointer);
var                                     {MINUS}
  last1, last2: val_c;
begin
  put_funcop(uminus);
  last1:= get_valifo(p_valifo(valifo));
  last2:= get_valifo(p_valifo(valifo));
  if ((last1 <> vnum) or (last2 <> vnum)) then
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure loeq_op(var valifo: pointer); {<=}
var
  last1, last2: val_c;
begin
  last1:= get_valifo(p_valifo(valifo));
  last2:= get_valifo(p_valifo(valifo));
  if ((last1 = vnum) and (last2 = vnum)) then
    put_funcop(uloeq)
  else if ((last1 = vstr) and (last2 = vstr)) then
    put_funcop(usloeq)
  else
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure hieq_op(var valifo: pointer); {>=}
var
  last1, last2: val_c;
begin
  last1:= get_valifo(p_valifo(valifo));
  last2:= get_valifo(p_valifo(valifo));
  if ((last1 = vnum) and (last2 = vnum)) then
    put_funcop(uhieq)
  else if ((last1 = vstr) and (last2 = vstr)) then
    put_funcop(ushieq)
  else
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure lo_op(var valifo: pointer);   {<}
var
  last1, last2: val_c;
begin
  last1:= get_valifo(p_valifo(valifo));
  last2:= get_valifo(p_valifo(valifo));
  if ((last1 = vnum) and (last2 = vnum)) then
    put_funcop(ulo)
  else if ((last1 = vstr) and (last2 = vstr)) then
    put_funcop(uslo)
  else
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure hi_op(var valifo: pointer);   {>}
var
  last1, last2: val_c;
begin
  last1:= get_valifo(p_valifo(valifo));
  last2:= get_valifo(p_valifo(valifo));
  if ((last1 = vnum) and (last2 = vnum)) then
    put_funcop(uhi)
  else if ((last1 = vstr) and (last2 = vstr)) then
    put_funcop(ushi)
  else
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure eq_op(var valifo: pointer);   {==}
var
  last1, last2: val_c;
begin
  last1:= get_valifo(p_valifo(valifo));
  last2:= get_valifo(p_valifo(valifo));
  if ((last1 = vnum) and (last2 = vnum)) then
    put_funcop(ueq)
  else if ((last1 = vstr) and (last2 = vstr)) then
    put_funcop(useq)
  else
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure neq_op(var valifo: pointer);  {!=}
var
  last1, last2: val_c;
begin
  last1:= get_valifo(p_valifo(valifo));
  last2:= get_valifo(p_valifo(valifo));
  if ((last1 = vnum) and (last2 = vnum)) then
    put_funcop(uneq)
  else if ((last1 = vstr) and (last2 = vstr)) then
    put_funcop(usneq)
  else
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure and_op(var valifo: pointer);  {&&}
var
  last1, last2: val_c;
begin
  put_funcop(uand);
  last1:= get_valifo(p_valifo(valifo));
  last2:= get_valifo(p_valifo(valifo));
  if ((last1 <> vnum) or (last2 <> vnum)) then
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

procedure or_op(var valifo: pointer);   {||}
var
  last1, last2: val_c;
begin
  put_funcop(uor);
  last1:= get_valifo(p_valifo(valifo));
  last2:= get_valifo(p_valifo(valifo));
  if ((last1 <> vnum) or (last2 <> vnum)) then
    error(comp, c_type, line, col);
  add2valifo(p_valifo(valifo), vnum);
end;

                (***** cbasic Command Procession *****)

procedure process_cmd;
var
  blockend: boolean;
  s: strlab;
  ch: char;
  ph: p_hash;
begin
  blockend:= true;
  eat_ws;
  if (lookch = '{') then begin
    blockend:= false;
    nextch;
    eat_ws;
  end;
  repeat
    inc(ifdistance);
    ch:= lookch;
    if (ch = '{') then
      process_cmd
    else if (ch = ';') then
      nextch                            {ignore semicolons (null commands)}
    else begin
      s:= gets(lab);
      ph:= srch_hash(cmdid + s);
      if (ph = NULL) then               {s is not command}
        error(comp, c_nocmd, line, col);
    end;
    eat_ws;
    if ((lookch = '}') and (blockend = false)) then begin
      nextch;                           {it's block of this procerure}
      blockend:= true;                  {end of block reached!}
    end;
  until (blockend);
end;

               (***** cbasic Command Procedures *****)

procedure check_semicolon;
begin
  eat_ws;
  if (getch <> ';') then
    error(comp, c_semic, line, col);
end;

procedure put_ucmd(uproc: ucmd_c);
var
  uc: t_ucmd;
begin
  uc.line:= line;
  uc.col:= col;
  uc.ucmd:= uproc;
  put_upgm(@uc, sizeof(uc));
end;

procedure process_forward(p: p_hforw);
var                                     {put everywhere before actual offset}
  l: long;
  q: p_hforw;
begin
  l:= uprogram_top;
  while (p <> NULL) do begin
    mov_upgm(p^.ofs, @l, sizeof(l));
    q:= p^.next;
    mfree(pointer(p));
    p:= q;
  end;
end;

procedure process_cforw(level: long; var p: p_cforw);
var                                     {put everywhere before actual offset}
  l: long;
  q: p_cforw;
begin
  l:= uprogram_top;
  while (p <> NULL) do begin
    if (p^.level = level) then
      mov_upgm(p^.ofs, @l, sizeof(l));
    q:= p^.next;
    mfree(pointer(p));
    p:= q;
  end;
end;

procedure main_cmd;                     {MAIN}
begin
  if (uprogram_start <> maxlong) then
    error(comp, c_2lab, line, col);
  uprogram_start:= uprogram_top;
  process_cmd;
  put_ucmd(okend);
end;

procedure label_cmd;                    {LABEL}
var
  s: strlab;
  ph: p_hash;
begin
  s:= clabid + gets(lab);
  ph:= srch_hash(s);
  if (ph <> NULL) then begin            {already referenced}
    if (ph^.clab.flag = forw) then
      process_forward(ph^.clab.forw)
    else
      error(comp, c_2lab, line, col);   {defined elsewhere}
    ph^.clab.flag:= done;
    ph^.clab.ofs:= uprogram_top;
  end
  else begin
    ph:= malloc(sizeof(t_hash));        {not defined yet -- add 2 hash table}
    ph^.name:= s;
    ph^.clab.flag:= done;
    ph^.clab.ofs:= uprogram_top;
    add2hash(ph);
  end;
  check_semicolon;
end;

procedure goto_cmd;                     {GOTO}
var
  s: strlab;
  ph: p_hash;
  l: long;
  pf: p_hforw;
begin
  put_ucmd(jump);
  s:= clabid + gets(lab);
  ph:= srch_hash(s);
  if (ph <> NULL) then begin            {defined before}
    if (ph^.clab.flag = done) then
      l:= ph^.clab.ofs
    else begin                          {referenced before}
      pf:= malloc(sizeof(t_hforw));
      pf^.next:= ph^.clab.forw^.next;
      pf^.ofs:= uprogram_top;
      ph^.clab.forw:= pf;               {add to chain}
    end;
  end
  else begin                            {not referenced nor defined}
    pf:= malloc(sizeof(t_hforw));
    pf^.next:= NULL;
    pf^.ofs:= uprogram_top;
    ph:= malloc(sizeof(t_hash));
    ph^.name:= s;
    ph^.clab.flag:= forw;
    ph^.clab.forw:= pf;
    add2hash(ph);
  end;
  put_upgm(@l, sizeof(l));              {store offset / make space only}
  check_semicolon;
end;

procedure gosub_cmd;                    {GOSUB}
var
  s: strlab;
  ph: p_hash;
  l: long;
  pf: p_hforw;
begin
  put_ucmd(call);
  s:= clabid + gets(lab);
  ph:= srch_hash(s);
  if (ph <> NULL) then begin            {defined before}
    if (ph^.clab.flag = done) then
      l:= ph^.clab.ofs
    else begin                          {referenced before}
      pf:= malloc(sizeof(t_hforw));
      pf^.next:= ph^.clab.forw^.next;
      pf^.ofs:= uprogram_top;
      ph^.clab.forw:= pf;               {add to chain}
    end;
  end
  else begin                            {not referenced nor defined}
    pf:= malloc(sizeof(t_hforw));
    pf^.next:= NULL;
    pf^.ofs:= uprogram_top;
    ph:= malloc(sizeof(t_hash));
    ph^.name:= s;
    ph^.clab.flag:= forw;
    ph^.clab.forw:= pf;
    add2hash(ph);
  end;
  put_upgm(@l, sizeof(l));              {store offset / make space only}
end;

procedure sub_cmd;                      {SUB}
var
  s: strlab;
  ph: p_hash;
begin
  s:= clabid + gets(lab);
  ph:= srch_hash(s);
  if (ph <> NULL) then begin            {already referenced}
    if (ph^.clab.flag = forw) then
      process_forward(ph^.clab.forw)
    else
      error(comp, c_2lab, line, col);   {defined elsewhere}
    ph^.clab.flag:= done;
    ph^.clab.ofs:= uprogram_top;
  end
  else begin
    ph:= malloc(sizeof(t_hash));        {not defined yet -- add 2 hash table}
    ph^.name:= s;
    ph^.clab.flag:= done;
    ph^.clab.ofs:= uprogram_top;
    add2hash(ph);
  end;
  process_cmd;
  put_ucmd(ret);
end;

procedure return_cmd;                   {RETURN}
begin
  put_ucmd(ret);
  check_semicolon;
end;

procedure let_cmd;                      {LET}
var
  letype: val_c;
begin
  put_ucmd(move);
  letype:= process_expr(variable);
  if (letype = vempty) then
    error(comp, c_assign, line, col);
  eat_ws;
  if (getch <> '=') then
    error(comp, c_assign, line, col);
  if (process_expr(expression) <> letype) then
    error(comp, c_expr, line, col);
  check_semicolon;
end;

procedure dim_cmd;                      {DIM}
var
  pc: p_char;
  back: long;
begin
  put_ucmd(newa);
  back:= uprogram_top;
  if (process_expr(variable) = vempty) then
    error(comp, c_array, line, col);
  pc:= uprogram;
  inc(pc, back);
  if ((p_uexpr(pc)^.flags and ue_array) = 0) then
    error(comp, c_arrex, line, col);
  check_semicolon;
end;

procedure print_cmd;                    {PRINT}
var
  i: int;
  terminate: boolean;
  back: long;
begin
  put_ucmd(output);
  back:= uprogram_top;
  put_upgm(@i, sizeof(i));              {make space only}
  i:= 0;
  terminate:= false;
  repeat
    if (process_expr(expression) = vempty) then
      error(comp, c_expr, line, col);
    eat_ws;
    if (lookch = ',') then
      nextch
    else
      terminate:= true;
    inc(i);
  until (terminate);
  mov_upgm(back, @i, sizeof(i));        {store number of expr. to print}
  check_semicolon;
end;

procedure input_cmd;                    {INPUT}
var
  i: int;
  terminate: boolean;
  back: long;
begin
  put_ucmd(input);
  eat_ws;
  back:= uprogram_top;
  put_upgm(@i, sizeof(i));              {make space only}
  i:= 0;
  terminate:= false;
  repeat
    if (process_expr(variable) = vempty) then
      error(comp, c_comma, line, col);
    eat_ws;
    if (lookch = ',') then
      nextch
    else
      terminate:= true;
    inc(i);
  until (terminate);
  mov_upgm(back, @i, sizeof(i));        {store number of expr. to print}
  check_semicolon;
end;

procedure data_cmd;                     {DATA}
var
  pc: p_char;
  back: long;
  l: long;
  s: strlab;
  ph: p_hash;
  terminate: boolean;
begin
  put_ucmd(jump);
  back:= uprogram_top;
  put_upgm(@l, sizeof(l));              {make space only}

  s:= dlabid + gets(lab);
  ph:= srch_hash(s);
  if (ph <> NULL) then begin            {already referenced}
    if (ph^.dlab.flag = forw) then
      process_forward(ph^.dlab.forw)
    else
      error(comp, c_2lab, line, col);   {defined elsewhere}
    ph^.dlab.flag:= done;
    ph^.dlab.ofs:= uprogram_top;
  end
  else begin
    ph:= malloc(sizeof(t_hash));        {not defined yet -- add 2 hash table}
    ph^.name:= s;
    ph^.dlab.flag:= done;
    ph^.dlab.ofs:= uprogram_top;
    add2hash(ph);
  end;

  terminate:= false;
  repeat
    if (udata_start <> maxlong) then begin
      l:= uprogram_top;                 {previous chain entry}
      mov_upgm(udata_top, @l, sizeof(l));
      udata_top:= uprogram_top;
    end else begin
      udata_top:= uprogram_top;         {first data}
      udata_start:= udata_top;
    end;
    l:= 0;                              {0 = end of chain}
    put_upgm(@l, sizeof(l));            {make space only}
    if (process_expr(expression) = vempty) then
      error(comp, c_expr, line, col);
    eat_ws;
    if (lookch = ',') then
      nextch
    else
      terminate:= true;
  until (terminate);
  l:= uprogram_top;
  mov_upgm(back, @l, sizeof(l));
  check_semicolon;
end;

procedure read_cmd;                     {READ}
var
  back: long;
  i: int;
  terminate: boolean;
begin
  put_ucmd(readv);
  back:= uprogram_top;
  put_upgm(@i, sizeof(i));              {make space only}
  i:= 0;
  terminate:= false;
  repeat
    if (process_expr(variable) = vempty) then
      error(comp, c_expr, line, col);
    eat_ws;
    if (lookch = ',') then
      nextch
    else
      terminate:= true;
    inc(i);
  until (terminate);
  mov_upgm(back, @i, sizeof(i));        {store # of expr. to read}
  check_semicolon;
end;

procedure restore_cmd;                  {RESTORE}
var
  pf: p_hforw;
  ph: p_hash;
  s: strlab;
  l: long;
begin
  put_ucmd(restore);
  s:= dlabid + gets(lab);
  ph:= srch_hash(s);
  if (ph <> NULL) then begin            {defined before}
    if (ph^.dlab.flag = done) then
      l:= ph^.dlab.ofs
    else begin                          {referenced before}
      pf:= malloc(sizeof(t_hforw));
      pf^.next:= ph^.dlab.forw^.next;
      pf^.ofs:= uprogram_top;
      ph^.dlab.forw:= pf;               {add to chain}
    end;
  end
  else begin                            {not referenced nor defined}
    pf:= malloc(sizeof(t_hforw));
    pf^.next:= NULL;
    pf^.ofs:= uprogram_top;
    ph:= malloc(sizeof(t_hash));
    ph^.name:= s;
    ph^.dlab.flag:= forw;
    ph^.dlab.forw:= pf;
    add2hash(ph);
  end;
  put_upgm(@l, sizeof(l));              {store offset / make space only}
end;

procedure if_cmd;                       {IF}
var
  l: long;
  back1: long;
begin
  put_ucmd(jmpf);
  back1:= uprogram_top;
  put_upgm(@l, sizeof(l));              {make space only}
  check_lpar;                           {'('}
  if(process_expr(expression) <> vnum) then
    error(comp, c_nexpex, line, col);
  check_rpar;                           {')'}
  process_cmd;
  l:= uprogram_top;
  mov_upgm(back1, @l, sizeof(l));       {if not true, jump here}
  mov_upgm(l, @back1, sizeof(back1));   {store offset for possibly else}
  ifdistance:= 0;
end;

procedure else_cmd;                     {ELSE}
var
  pc: p_char;
  ifofs, back2: long;
  l: long;
begin
  if (ifdistance > 1) then
    error(comp, c_else, line, col);
  pc:= uprogram;
  inc(pc, uprogram_top);
  ifofs:= p_long(pc)^;                  {get offset of previous if}
  put_ucmd(jump);
  back2:= uprogram_top;
  put_upgm(@l, sizeof(l));              {make space only}
  l:= uprogram_top;
  mov_upgm(ifofs, @l, sizeof(l));       {store right jmpf offset}
  process_cmd;
  l:= uprogram_top;
  mov_upgm(back2, @l, sizeof(l));       {store right 'skip else' jump offset}
end;

procedure for_cmd;                      {FOR}
var
  top1, top2: long;
  back3, back4: long;
  l: long;
begin
  inc(cyclelevel);
  check_lpar;                           {'('}
  process_cmd;                          {cmd1}
  top1:= uprogram_top;
  put_ucmd(jmpt);
  back3:= uprogram_top;
  put_upgm(@l, sizeof(l));              {make space only}
  if (process_expr(expression) <> vnum) then
    error(comp, c_nexpex, line, col);
  check_semicolon;                      {expr}
  put_ucmd(jump);
  back4:= uprogram_top;
  put_upgm(@l, sizeof(l));
  top2:= uprogram_top;
  process_cmd;                          {cmd2}
  put_ucmd(jump);
  put_upgm(@top1, sizeof(top1));
  check_rpar;                           {')'}
  l:= uprogram_top;                     {do not put uprogram_top directly!}
  mov_upgm(back3, @l, sizeof(l));
  process_cmd;                          {cmd3}
  process_cforw(cyclelevel, cont_forwp);{continue goes here}
  put_ucmd(jump);
  put_upgm(@top2, sizeof(top2));
  l:= uprogram_top;                     {do not put uprogram_top directly!}
  mov_upgm(back4, @l, sizeof(l));
  process_cforw(cyclelevel, break_forwp);
  dec(cyclelevel);                      {break goes here}
end;

procedure while_cmd;                    {WHILE}
var
  top1, back2: long;
  l: long;
begin
  inc(cyclelevel);
  check_lpar;                           {'('}
  top1:= uprogram_top;
  put_ucmd(jmpf);
  back2:= uprogram_top;
  put_upgm(@l, sizeof(l));              {make space only}
  if (process_expr(expression) <> vnum) then
    error(comp, c_nexpex, line, col);
  check_rpar;                           {')'}
  process_cmd;
  process_cforw(cyclelevel, cont_forwp);{continue goes here}
  put_ucmd(jump);
  put_upgm(@top1, sizeof(top1));
  l:= uprogram_top;                     {do not put uprogram_top directly!}
  mov_upgm(back2, @l, sizeof(l));
  process_cforw(cyclelevel, break_forwp);
  dec(cyclelevel);                      {break goes here}
end;

procedure do_cmd;                       {DO}
var
  top1, back2: long;
  l: long;
  s: strlab;
begin
  inc(cyclelevel);
  top1:= uprogram_top;
  process_cmd;                          {cmd}
  if (upstr(gets(lab)) <> 'WHILE') then
    error(comp, c_while, line, col);
  check_lpar;                           {'('}
  process_cforw(cyclelevel, cont_forwp);{continue goes here}
  put_ucmd(jmpt);
  put_upgm(@top1, sizeof(top1));
  if (process_expr(expression) <> vnum) then
    error(comp, c_nexpex, line, col);
  check_rpar;                           {')'}
  process_cforw(cyclelevel, break_forwp);
  dec(cyclelevel);                      {break goes here}
  check_semicolon;
end;

procedure break_cmd;                    {BREAK}
var
  pf: p_cforw;
  l: long;
begin
  put_ucmd(jump);
  l:= uprogram_top;
  put_upgm(@l, sizeof(l));              {make space only}
  pf:= cnew(sizeof(t_cforw));
  pf^.next:= break_forwp;
  pf^.level:= cyclelevel;
  pf^.ofs:= l;
  break_forwp:= pf;
end;

procedure continue_cmd;                 {CONTINUE}
var
  pf: p_cforw;
  l: long;
begin
  put_ucmd(jump);
  l:= uprogram_top;
  put_upgm(@l, sizeof(l));              {make space only}
  pf:= cnew(sizeof(t_cforw));
  pf^.next:= cont_forwp;
  pf^.level:= cyclelevel;
  pf^.ofs:= l;
  cont_forwp:= pf;
end;

procedure at_cmd;                       {AT}
begin
  put_ucmd(at);
  if (process_expr(expression) <> vnum) then
    error(comp, c_expr, line, col);
  eat_ws;
  check_comma;
  if (process_expr(expression) <> vnum) then
    error(comp, c_expr, line, col);
  check_semicolon;
end;

procedure tab_cmd;                      {TAB}
begin
  put_ucmd(tab);
  if (process_expr(expression) <> vnum) then
    error(comp, c_expr, line, col);
  check_semicolon;
end;

procedure cls_cmd;                      {CLS}
begin
  put_ucmd(cls);
  check_semicolon;
end;

procedure randomize_cmd;                {RANDOMIZE}
begin
  put_ucmd(randomi);
  check_semicolon;
end;

procedure stop_cmd;                     {STOP}
begin
  put_ucmd(stop);
  check_semicolon;
end;


                   (***** Main Compiler Procedure *****)

procedure compiler;
var
  xpl: ^long;
  ch: char;
begin
  init_hash;                            {initialize hash table}

  fbuffer:= malloc(fbuffer_size);       {src file buffer}
  if (fbuffer = NULL) then
    error(comp, c_lowmem, line, col);
  fbuffer_max:= 0;
  fbuffer_top:= 0;
  ch:= getch; ch:= getch;               {initialize chact, chnext}
  line:= 1; col:= 1;

  uprogram:= malloc(compiled_size);     {uprogram}
  if (uprogram = NULL) then
    error(comp, c_lowmem, 0, 0);

  uvars:= malloc(variable_size);        {uvars}
  if (uvars = NULL) then
    error(comp, c_lowmem, 0, 0);

  uprogram_top:= 0;
  uprogram_start:= maxlong;
  udata_top:= 0;
  udata_start:= maxlong;
  uvars_top:= 0;

  cont_forwp:= NULL;
  break_forwp:= NULL;

  ifdistance:= 2;
  cyclelevel:= 0;
  repeat
    process_cmd;
    eat_ws;
  until (lookch = EOFILE);

  if (uprogram_start = maxlong) then
    uprogram_start:= 0;

  put_ucmd(okend);

  mfree(pointer(fbuffer));              {free unused parts of alloc. memory}
  uprogram:= realloc(uprogram, uprogram_top);
  uvars:= realloc(uvars, uvars_top);

  free_hash;
end;

       (******************** INTERPRETER ********************)

               (***** Interpreter Constants *****)

          (***** Miscellaenous Interpreter Procedures *****)

procedure get_upgm(p: pointer; len: long);
var                                     {get something from upgm}
  li: long;
  px: ^long;
  psrc, pdst: ^char;
begin
  psrc:= uprogram;
  pdst:= p;
  inc(psrc, uprogramp);
  uprogramp:= uprogramp + len;
  for li:= 1 to len do begin
    pdst^:= psrc^;
    inc(psrc);
    inc(pdst);
  end;
end;

procedure destroy_str(s: t_uexpr);      {free string when necessary}
begin
  if ((s.flags and ue_temp) <> 0) then
    mfree(pointer(s.s_eval));
end;

                       (***** Stack Procedures *****)

procedure push(e: t_uexpr);             {expression stack procedures}
var
  xp: p_uexpr;
  pl: ^long;
  xl: long;
begin
  pl:= pointer(ustack);
  dec(pl, 1);
  xl:= pl^;
  if ((ustackp * sizeof(t_uexpr)) >= pl^) then begin
    ustack:= realloc(pointer(ustack),
      (ustackp + stack_plus_size) * sizeof(t_uexpr));
    if (ustack = NULL) then
      error(inter, i_stacko, line, col);
  end;
  xp:= pointer(ustack);
  inc(xp, ustackp);
  xp^:= e;
  inc(ustackp);
end;

procedure pop(var e: t_uexpr);
var
  xp: p_uexpr;
begin
  if (ustackp = 0) then                 {stack underflow}
    error(inter, i_stacku, line, col);
  xp:= pointer(ustack);
  dec(ustackp);
  inc(xp, ustackp);
  e:= xp^;
end;

             (***** Evaluation of uExpressions (uVars) *****)

procedure eval_uexpr(var uexpr: t_uexpr);
var
  ue: t_uexpr;
  xpc: ^char;
  dims: int;
  i: int;
  l: long;
  array_ofs: long;
  isstr: boolean;
  adim, sdim: int;
  fg: short;
begin
  repeat
    xpc:= uprogram;                     {faster than calling get_upgm}
    inc(xpc, uprogramp);
    inc(uprogramp, sizeof(t_uexpr));
    ue:= p_uexpr(xpc)^;

    fg:= ue.flags;
    if ((fg and ue_funcop) <> 0) then
       xxx
    else begin
      isstr:= ((fg and ue_string) <> 0);
      if ((fg and ue_array) <> 0) then begin
        get_upgm(@dims, sizeof(dims));  {ARRAY}
        xpc:= uvars;
        if (isstr) then
          inc(xpc, ue.s_array)
        else
          inc(xpc, ue.n_array);
        if (p_array(xpc)^.dims <> dims) then
          if (p_array(xpc)^.dims = 0) then
            error(inter, i_uarray, line, col)
          else
            error(inter, i_dims, line, col);
        xpc:= p_array(xpc)^.ptr;       {shouldn't be NULL}
        for i:= 1 to dims do begin
          eval_uexpr(ue);
          if (not do2int(ue)) then
            error(inter, i_dims, line, col);
          adim:= ue.n_eval.l;           {actual dim}
          sdim:= p_int(xpc)^;           {stored one}
          inc(p_int(xpc));
          if ((adim <= 0) or (adim > sdim)) then
            error(inter, i_dims, line, col);
          if (i > 1) then
            array_ofs:= array_ofs*sdim+(adim-1)
          else
            array_ofs:= adim-1;         {first time}
        end;
        if (isstr) then begin           {store result as evaluated}
          array_ofs:= array_ofs * sizeof(p_string);
          inc(xpc, array_ofs);
          if (pp_string(xpc)^ <> NULL) then
            ue.s_eval:= pp_string(xpc)^
          else
            ue.s_eval:= p_string(@nullstr);
          ue.flags:= ue_evaled or ue_string;
        end
        else begin
          array_ofs:= array_ofs * sizeof(t_uexpr);
          inc(xpc, array_ofs);
          ue:= p_uexpr(xpc)^;
          ue.flags:= ue.flags or ue_evaled;
        end;
      end                               {constant/variable}
      else if ((fg and ue_const) <> 0) then begin
        if (isstr) then begin           
          xpc:= uvars;                  {string CONSTANT}
          inc(xpc, ue.s_const);
          ue.s_eval:= p_string(xpc);    {store pointer}
          ue.flags:= ue_evaled or ue_string;
        end else                        {numeric CONSTANT (do nothing)}
          ue.flags:= ue.flags or ue_evaled;
      end
      else begin
        if (isstr) then begin           {string VARIABLE}
          xpc:= uvars;
          inc(xpc, ue.s_var);
          if (pp_string(xpc)^ <> NULL) then
            ue.s_eval:= pp_string(xpc)^
          else
            ue.s_eval:= p_string(@nullstr);
          ue.flags:= ue_evaled or ue_string;
        end
        else begin                      {numeric VARIABLE}
          xpc:= uvars;
          inc(xpc, ue.n_var);
          ue:= p_uexpr(xpc)^;
          ue.flags:= ue.flags or ue_evaled;
        end;
      end;
      push(ue);
    end;
  until ((fg and ue_last) <> 0);
  pop(uexpr);
end;

procedure eval_uvar(var ue: t_uexpr);
var
  xpc: ^char;
  dims: int;
  i: int;
  l: long;
  array_ofs: long;
  isstr: boolean;
  adim, sdim: int;
  fg: short;
begin
  xpc:= uprogram;
  inc(xpc, uprogramp);
  inc(uprogramp, sizeof(t_uexpr));
  ue:= p_uexpr(xpc)^;
  fg:= ue.flags;
  if ((fg and ue_funcop) <> 0) then begin
    pop(ue);
  end
  else begin
    isstr:= ((fg and ue_string) <> 0);
    if ((fg and ue_array) <> 0) then begin
      get_upgm(@dims, sizeof(dims));    {ARRAY}
      xpc:= uvars;
      if (isstr) then
        inc(xpc, ue.s_array)
      else
        inc(xpc, ue.n_array);
      if (p_array(xpc)^.dims <> dims) then
        if (p_array(xpc)^.dims = 0) then
          error(inter, i_uarray, line, col)
        else
          error(inter, i_dims, line, col);
      xpc:= p_array(xpc)^.ptr;
      for i:= 1 to dims do begin
        eval_uexpr(ue);
        if (not do2int(ue)) then
          error(inter, i_dims, line, col);
        adim:= ue.n_eval.l;             {actual dim}
        sdim:= p_int(xpc)^;             {stored one}
        inc(p_int(xpc));
        if ((adim <= 0) or (adim > sdim)) then
          error(inter, i_dims, line, col);
        if (i > 1) then
          array_ofs:= array_ofs*sdim+(adim-1)
        else
          array_ofs:= adim-1;           {first time}
      end;
      if (isstr) then begin             {store result as evaluated}
        array_ofs:= array_ofs * sizeof(p_string);
        inc(xpc, array_ofs);
        ue.s_varp:= pp_string(xpc);
        ue.flags:= ue_evaled or ue_string;
      end
      else begin
        array_ofs:= array_ofs * sizeof(t_uexpr);
        inc(xpc, array_ofs);
        ue.n_varp:= p_uexpr(xpc);
        ue.flags:= ue_evaled;
      end;
    end
    else if ((fg and ue_const) = 0) then begin
      if (isstr) then begin           {string VARIABLE}
        xpc:= uvars;
        inc(xpc, ue.s_var);
        ue.s_varp:= pp_string(xpc);
        ue.flags:= ue_evaled or ue_string;
      end
      else begin                      {numeric VARIABLE}
        xpc:= uvars;
        inc(xpc, ue.n_var);
        ue.n_varp:= p_uexpr(xpc);
        ue.flags:= ue_evaled;
      end;
    end else
      error(inter, i_wconst, line, col);
  end;
end;

          (***** uFunction (and uOperator) Procedures *****)

procedure push_bool(res: boolean);
var
  ue: t_uexpr;
begin
  ue.flags:= 0;
  if (res) then
    ue.n_eval.l:= 1
  else
    ue.n_eval.l:= 0;
  push(ue);
end;

function arcsin(fd: double): double;
var
  f2: double;
begin
  f2:= 1-fd*fd;
  if (f2 < 0) then
    error(inter, i_divz, line, col);
  arcsin:= arctan(fd/(sqr(f2)+1))*2;
end;

procedure sin_ufn;                      {SIN}
var
  ue: t_uexpr;
begin
  eval_uexpr(ue);
  int2float(ue);
  ue.n_eval.f:= sin(ue.n_eval.f);
  try2int(ue);
  push(ue);
end;

procedure cos_ufn;                      {COS}
var
  ue: t_uexpr;
begin
  eval_uexpr(ue);
  int2float(ue);
  ue.n_eval.f:= cos(ue.n_eval.f);
  try2int(ue);
  push(ue);
end;

procedure tan_ufn;                      {TAN}
var
  ue: t_uexpr;
  fd: double;
begin
  eval_uexpr(ue);
  int2float(ue);
  fd:= ue.n_eval.f;
  fd:= cos(fd);
  if (fd = 0) then
    error(inter, i_divz, line,  col);
  ue.n_eval.f:= sin(ue.n_eval.f) / fd;
  try2int(ue);
  push(ue);
end;

procedure asn_ufn;                      {ARCSIN}
var
  ue: t_uexpr;
begin
  eval_uexpr(ue);
  int2float(ue);
  ue.n_eval.f:= arcsin(ue.n_eval.f);
  try2int(ue);
  push(ue);
end;

procedure acs_ufn;                      {ARCCOS}
var
  ue: t_uexpr;
begin
  eval_uexpr(ue);
  int2float(ue);
  ue.n_eval.f:= pi/2 - arcsin(ue.n_eval.f);
  try2int(ue);
  push(ue);
end;

procedure atn_ufn;                      {ARCTAN}
var
  ue: t_uexpr;
begin
  eval_uexpr(ue);
  int2float(ue);
  ue.n_eval.f:= arctan(ue.n_eval.f);
  try2int(ue);
  push(ue);
end;

procedure sqr_ufn;                      {SQUARE ROOT}
var
  ue: t_uexpr;
begin
  eval_uexpr(ue);
  int2float(ue);
  if (ue.n_eval.f < 0) then
    error(inter, i_divz, line, col);
  ue.n_eval.f:= sqr(ue.n_eval.f);
  try2int(ue);
  push(ue);
end;

procedure ln_ufn;                       {LN_e}
var
  ue: t_uexpr;
begin
  eval_uexpr(ue);
  int2float(ue);
  if (ue.n_eval.f <= 0) then
    error(inter, i_divz, line, col);
  ue.n_eval.f:= ln(ue.n_eval.f);
  try2int(ue);
  push(ue);
end;

procedure exp_ufn;                      {EXP}
var
  ue: t_uexpr;
begin
  eval_uexpr(ue);
  int2float(ue);
  ue.n_eval.f:= exp(ue.n_eval.f);
  try2int(ue);
  push(ue);
end;

procedure sgn_ufn;                      {SIGNUM}
var
  ue: t_uexpr;
  f: float;
  l: long;
begin
  eval_uexpr(ue);
  int2float(ue);
  f:= ue.n_eval.f;
  if (f < 0) then
    l:= 1
  else if (f > 0) then
    l:= 1
  else
    l:= 0;
  ue.flags:= ue.flags xor ue_float;     {return int}
  ue.n_eval.l:= l;
  push(ue);
end;

procedure int_ufn;                      {INT}
var
  ue: t_uexpr;
  xf: float;
begin
  eval_uexpr(ue);
  if (not(do2int(ue))) then
    error(inter, i_conv, line, col);
  push(ue);
end;

procedure abs_ufn;                      {ABS}
var
  ue: t_uexpr;
begin
  eval_uexpr(ue);
  if ((ue.flags and ue_float) <> 0) then begin
    if (ue.n_eval.f < 0) then
      ue.n_eval.f:= ue.n_eval.f;
  end else
    if (ue.n_eval.l < 0) then
      ue.n_eval.l:= -ue.n_eval.l;
  push(ue);
end;

procedure rnd_ufn;                      {RND}
var
  ue: t_uexpr;
begin
  ue.flags:= ue_evaled or ue_float;
  ue.n_eval.f:= random(maxint) / maxint;
  push(ue);
end;

procedure chrs_ufn;                     {CHR$}
var
  ue: t_uexpr;
  ps: p_string;
  l: long;
begin
  eval_uexpr(ue);
  if (not do2int(ue)) then
    error(inter, i_char, line, col);
  l:= ue.n_eval.l;
  if ((l < 0) or (l > 255)) then
    error(inter, i_char, line, col);
  ps:= malloc(1+1);                     {one char}
  if (ps = NULL) then
    error(inter, i_lowmem, line, col);
  ps^:= chr(l);
  ue.flags:= ue_evaled or ue_string or ue_temp;
  ue.s_eval:= ps;
  push(ue);
end;

procedure idxs_ufn;                     {IDX$}
var
  ue1, ue2: t_uexpr;
  ps: p_string;
  l: long;
begin
  eval_uexpr(ue1);                      {string}
  eval_uexpr(ue2);                      {number}
  if (not do2int(ue2)) then
    error(inter, i_char, line, col);
  l:= ue2.n_eval.l;
  if (l > length(ue1.s_eval^)) then
    error(inter, i_index, line, col);
  ps:= malloc(1+1);                     {one char}
  if (ps = NULL) then
    error(inter, i_lowmem, line, col);
  ps^:= ue1.s_eval^[l];
  ue2.flags:= ue_evaled or ue_string or ue_temp;
  ue2.s_eval:= ps;
  push(ue2);
  destroy_str(ue1);
end;

procedure vidxs_ufn;                    {v-IDX$}
var
  ue1, ue2, ue3: t_uexpr;
  ps: p_string;
  ix, l1, i: int;
begin
  eval_uvar(ue1);                       {string var}
  eval_uexpr(ue2);                      {number}
  if (not do2int(ue2)) then
    error(inter, i_index, line, col);
  ix:= ue2.n_eval.l;                    {ix = index in ue1}
  if (ix <= 0) then
    error(inter, i_index, line, col);
  ps:= ue1.s_varp^;
  if (ps <> NULL) then
    l1:= length(ps^)
  else
    l1:= 0;
  if (ix > l1) then begin               {index out of string}
    if (ps <> NULL) then
      ps:= realloc(pointer(ps), ix+1)
    else begin
      ps:= malloc(ix+1);
      ps^:= '';
    end;
    if (ps = NULL) then
      error(inter, i_lowmem, line, col);
    for i:= l1+1 to ix do
      ps^:= ps^ + ' ';
    ue1.s_varp^:= ps;
  end;
  ue1.c_varp:= @ps^[ix];
  ue1.flags:= ue_evaled or ue_char or ue_string;
  push(ue1);
end;

procedure len_ufn;                      {LEN}
var
  ue1, ue2: t_uexpr;
begin
  eval_uexpr(ue1);
  ue2.flags:= ue_evaled;
  ue2.n_const.l:= length(ue1.s_eval^);
  push(ue2);
  destroy_str(ue1);
end;

procedure uminus_uop;                   {-X}
var
  ue: t_uexpr;
begin
  pop(ue);
  if ((ue.flags and ue_float) <> 0) then
    ue.n_eval.f:= -ue.n_eval.f
  else
    ue.n_eval.l:= -ue.n_eval.l;
  push(ue);
end;

procedure neg_uop;                      {!X}
var
  ue: t_uexpr;
begin
  pop(ue);
  if (iszero(ue)) then
    ue.n_eval.l:= 1
  else
    ue.n_eval.l:= 0;
  ue.flags:= ue_evaled;
  push(ue);
end;

procedure times_uop;                    {X * Y}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  if (both_ints(ue1, ue2)) then
    ue1.n_eval.l:= ue1.n_eval.l * ue2.n_eval.l
  else begin
    int2float(ue1);
    int2float(ue2);
    ue1.n_eval.f:= ue1.n_eval.f * ue2.n_eval.f;
    try2int(ue1);
  end;
  push(ue1);
end;

procedure div_uop;                      {X / Y}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  if (iszero(ue2)) then
    error(inter, i_divz, line, col);
  if (both_ints(ue1, ue2)) then
    ue1.n_eval.l:= ue1.n_eval.l div ue2.n_eval.l
  else begin
    int2float(ue1);
    int2float(ue2);
    ue1.n_eval.f:= ue1.n_eval.f / ue2.n_eval.f;
    try2int(ue1);
  end;
  push(ue1);
end;

procedure mod_uop;                      {X % Y}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  if (iszero(ue2)) then
    error(inter, i_divz, line, col);
  if (both_ints(ue1, ue2)) then
    ue1.n_eval.l:= ue1.n_eval.l mod ue2.n_eval.l
  else
    error(inter, i_divz, line, col);
  push(ue1);
end;

procedure plus_uop;                     {PLUS}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  if (both_ints(ue1, ue2)) then
    ue1.n_eval.l:= ue1.n_eval.l + ue2.n_eval.l
  else begin
    int2float(ue1);
    int2float(ue2);
    ue1.n_eval.f:= ue1.n_eval.f + ue2.n_eval.f;
    try2int(ue1);
  end;
  push(ue1);
end;

procedure minus_uop;                    {MINUS}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  if (both_ints(ue1, ue2)) then
    ue1.n_eval.l:= ue1.n_eval.l - ue2.n_eval.l
  else begin
    int2float(ue1);
    int2float(ue2);
    ue1.n_eval.f:= ue1.n_eval.f - ue2.n_eval.f;
    try2int(ue1);
  end;
  push(ue1);
end;

procedure loeq_uop;                     {x <= x}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  if (both_ints(ue1, ue2)) then
    push_bool(ue1.n_eval.l <= ue2.n_eval.l)
  else begin
    int2float(ue1);
    int2float(ue2);
    push_bool(ue1.n_eval.f <= ue2.n_eval.f);
  end;
end;

procedure hieq_uop;                     {x >= x}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  if (both_ints(ue1, ue2)) then
    push_bool(ue1.n_eval.l >= ue2.n_eval.l)
  else begin
    int2float(ue1);
    int2float(ue2);
    push_bool(ue1.n_eval.f >= ue2.n_eval.f);
  end;
end;

procedure lo_uop;                       {x < x}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  if (both_ints(ue1, ue2)) then
    push_bool(ue1.n_eval.l < ue2.n_eval.l)
  else begin
    int2float(ue1);
    int2float(ue2);
    push_bool(ue1.n_eval.f < ue2.n_eval.f);
  end;
end;

procedure hi_uop;                       {x > x}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  if (both_ints(ue1, ue2)) then
    push_bool(ue1.n_eval.l > ue2.n_eval.l)
  else begin
    int2float(ue1);
    int2float(ue2);
    push_bool(ue1.n_eval.f > ue2.n_eval.f);
  end;
end;

procedure eq_uop;                       {x == x}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  if (both_ints(ue1, ue2)) then
    push_bool(ue1.n_eval.l = ue2.n_eval.l)
  else begin
    int2float(ue1);
    int2float(ue2);
    push_bool(ue1.n_eval.f = ue2.n_eval.f);
  end;
end;

procedure neq_uop;                      {x != x}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  if (both_ints(ue1, ue2)) then
    push_bool(ue1.n_eval.l <> ue2.n_eval.l)
  else begin
    int2float(ue1);
    int2float(ue2);
    push_bool(ue1.n_eval.f <> ue2.n_eval.f);
  end;
end;

procedure and_uop;                      {x && x}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  if (both_ints(ue1, ue2)) then
    push_bool((ue1.n_eval.l <> 0) and (ue2.n_eval.l <> 0))
  else begin
    int2float(ue1);
    int2float(ue2);
    push_bool((ue1.n_eval.f <> 0) and (ue2.n_eval.f <> 0));
  end;
end;

procedure or_uop;                       {x || x}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  if (both_ints(ue1, ue2)) then
    push_bool((ue1.n_eval.l <> 0) or (ue2.n_eval.l <> 0))
  else begin
    int2float(ue1);
    int2float(ue2);
    push_bool((ue1.n_eval.f <> 0) or (ue2.n_eval.f <> 0));
  end;
end;

procedure splus_uop;                    {s + s}
var
  ue1, ue2, ue: t_uexpr;
  ps: p_string;
begin
  pop(ue2); pop(ue1);
  ps:= malloc(length(ue1.s_eval^)+length(ue2.s_eval^)+1);
  if (ps = NULL) then
    error(inter, i_lowmem, line, col);
  ps^:= ue1.s_eval^+ue2.s_eval^;
  ue.flags:= ue_evaled or ue_string or ue_temp;
  ue.s_eval:= ps;
  push(ue);
  destroy_str(ue1);
  destroy_str(ue2);
end;

procedure sloeq_uop;                    {s <= s}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  push_bool(ue1.s_eval^ <= ue2.s_eval^);
  destroy_str(ue1);
  destroy_str(ue2);
end;

procedure shieq_uop;                    {s >= s}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  push_bool(ue1.s_eval^ >= ue2.s_eval^);
  destroy_str(ue1);
  destroy_str(ue2);
end;

procedure slo_uop;                      {s < s}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  push_bool(ue1.s_eval^ < ue2.s_eval^);
  destroy_str(ue1);
  destroy_str(ue2);
end;

procedure shi_uop;                      {s > s}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  push_bool(ue1.s_eval^ > ue2.s_eval^);
  destroy_str(ue1);
  destroy_str(ue2);
end;

procedure seq_uop;                      {s == s}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  push_bool(ue1.s_eval^ = ue2.s_eval^);
  destroy_str(ue1);
  destroy_str(ue2);
end;

procedure sneq_uop;                     {s != s}
var
  ue1, ue2: t_uexpr;
begin
  pop(ue2); pop(ue1);
  push_bool(ue1.s_eval^ <> ue2.s_eval^);
  destroy_str(ue1);
  destroy_str(ue2);
end;

                  (***** uCommand Procedures *****)

procedure load(ued, ues: p_uexpr);      {used by move and read}
var
  slen: integer;
  ps: p_string;
begin
  if ((ued^.flags and ue_string) <> 0) then begin
    if ((ued^.flags and ue_char) = 0) then begin
      slen:= length(ues^.s_eval^)+1;     {string}
      ps:= ued^.s_varp^;
      if (ps = NULL) then
        ps:= malloc(slen)
      else
        ps:= realloc(pointer(ps), slen);
      if (ps = NULL) then
        error(inter, i_lowmem, line, col);
      ps^:= ues^.s_eval^;
      ued^.s_varp^:= ps;
    end else
      ued^.c_varp^:= ues^.s_eval^[1];   {one character}
    destroy_str(ues^);
  end
  else begin
    ued^.n_varp^:= ues^;                {only float/int flag will stay}
    ued^.n_varp^.flags:= ued^.n_varp^.flags and ue_float;
  end;
end;

procedure jump_ucmd;                    {JUMP}
var
  l: long;
begin
  get_upgm(@l, sizeof(l));
  uprogramp:= l;
end;

procedure jmpt_ucmd;                    {JUMP when true}
var
  l: long;
  ue: t_uexpr;
begin
  get_upgm(@l, sizeof(l));
  eval_uexpr(ue);
  if (not iszero(ue)) then
    uprogramp:= l;
end;

procedure jmpf_ucmd;                    {JUMP when false}
var
  l: long;
  ue: t_uexpr;
begin
  get_upgm(@l, sizeof(l));
  eval_uexpr(ue);
  if (iszero(ue)) then
    uprogramp:= l;
end;

procedure call_ucmd;                    {CALL procedure}
var
  l: long;
  ue: t_uexpr;
begin
  ue.flags:= ue_retadr;
  ue.retadr:= uprogramp;
  push(ue);
  get_upgm(@l, sizeof(l));
  uprogramp:= l;
end;

procedure ret_ucmd;                     {RET from procedure}
var
  ue: t_uexpr;
begin
  pop(ue);
  if ((ue.flags and ue_retadr) = 0) then
    error(inter, i_smism, line, col);
  uprogramp:= ue.retadr;
end;

procedure move_ucmd;                    {MOVE}
var
  ue1, ue2: t_uexpr;
  ps: p_string;
  slen: int;
begin
  eval_uvar(ue1);
  eval_uexpr(ue2);                      {let ue1 = ue2}
  load(@ue1, @ue2);
end;

procedure newa_ucmd;                    {NEW ARRAY}
var
  ue, uea: t_uexpr;
  i, n: int;
  l: long;
  rsize, asize: long;
  pi: p_int;
  p: pointer;
  pa: p_array;
  isstr: boolean;
begin
  get_upgm(@uea, sizeof(uea));
  isstr:= ((uea.flags and ue_string) <> 0);

  pa:= uvars;                           {free possibly old array}
  if (isstr) then
    inc(p_char(pa), uea.s_array)
  else
    inc(p_char(pa), uea.n_array);
  if (pa^.dims <> 0) then
    mfree(pa^.ptr);

  get_upgm(@n, sizeof(n));
  p:= malloc(n * sizeof(int));
  if (p = NULL) then
    error(inter, i_lowmem, line, col);
  asize:= 1;
  pi:= p;
  for i:= 1 to n do begin               {multiple dimensions}
    eval_uexpr(ue);
    if (not do2int(ue)) then
      error(inter, i_dims, line, col);
    l:= ue.n_const.l;
    if ((l > maxint) or (l <= 0)) then
      error(inter, i_dims, line, col);
    asize:= asize * l;
    pi^:= int(l);
    inc(pi);
  end;

  if (isstr) then
    rsize:= sizeof(p_string)            {record size}
  else
    rsize:= sizeof(t_uexpr);
  p:= realloc(p, n * sizeof(int) + asize * rsize);
  if (p = NULL) then
    error(inter, i_lowmem, line, col);
  pa^.dims:= n;                         {update uvars}
  pa^.ptr:= p;

  inc(p_char(p), n * sizeof(int));      {initialize array}
  if (isstr) then
    for l:= 1 to asize do begin
      pp_string(p)^:= NULL;
      inc(pp_string(p));
    end
  else
    for l:= 1 to asize do begin
      p_uexpr(p)^.flags:= 0;
      p_uexpr(p)^.n_const.l:= 0;
      inc(p_uexpr(p));
    end;
end;

procedure read_ucmd;                    {READ}
var
  ue1, ue2: t_uexpr;
  i, n: int;
  oldp: long;
  slen: int;
  ps: p_string;
begin
  get_upgm(@n, sizeof(n));
  for i:= 1 to n do begin
    eval_uvar(ue1);                     {ue1 <- ue2}
    oldp:= uprogramp;
    if (udatap = 0) then
      error(inter, i_read, line, col);
    uprogramp:= udatap;
    get_upgm(@udatap, sizeof(udatap));  {read next data item offset}
    eval_uexpr(ue2);
    uprogramp:= oldp;

    if ((ue1.flags and ue_string) <>    {are both numbers/strings?}
        (ue2.flags and ue_string)) then
      error(inter, i_data, line, col);

    load(@ue1, @ue2);
  end;
end;

procedure restore_ucmd;                 {RESTORE}
begin
  get_upgm(@udatap, sizeof(udatap));
end;

procedure input_ucmd;                   {INPUT}
var
  ue, xue: t_uexpr;
  i, n, err: int;
  l: long;
  f: float;
  s: strmax;
  ps: p_string;
begin
  get_upgm(@n, sizeof(n));
  for i:= 1 to n do begin
    eval_uvar(ue);
    readln(s);
    if ((ue.flags and ue_string) <> 0) then begin
      ps:= ue.s_varp^;
      if (ps = NULL) then
        ps:= malloc(length(s)+1)
      else
        ps:= realloc(pointer(ps), length(s)+1);
      ps^:= s;
      ue.s_varp^:= ps;
    end
    else begin
      val(s, l, err);
      if (err = 0) then begin
        xue.flags:= ue_const;
        xue.n_const.l:= l;
      end
      else begin
        val(s, f, err);
        if (err = 0) then begin
          xue.flags:= ue_const or ue_float;
          xue.n_const.f:= f;
        end else
          error(inter, i_numf, line, col);
      end;
      ue.n_varp^:= xue;
    end
  end;
end;

procedure output_ucmd;                  {OUTPUT}
var
  ue: t_uexpr;
  i, n: int;
begin
  get_upgm(@n, sizeof(n));
  for i:= 1 to n do begin
    eval_uexpr(ue);
    if ((ue.flags and ue_string) <> 0) then begin
      write(ue.s_eval^);
      destroy_str(ue);
    end
    else if ((ue.flags and ue_float) <> 0) then
        write(ue.n_eval.f)
      else
        write(ue.n_eval.l);
  end;
  writeln;
end;

procedure at_ucmd;                      {AT}
var
  ue: t_uexpr;
  l, c: long;
begin
  eval_uexpr(ue);
  if (not do2int(ue)) then
    error(inter, i_param, line, col);
  l:= ue.n_eval.l;
  eval_uexpr(ue);
  if (not do2int(ue)) then
    error(inter, i_param, line, col);
  c:= ue.n_eval.l;
  if ((c <= 0) or (l <= 0)) then
    error(inter, i_param, line, col);
  write(ESC, '[', l, ';', c, 'f');      {ansi cursor move}
end;

procedure tab_ucmd;                     {TAB}
var
  ue: t_uexpr;
  c: long;
begin
  eval_uexpr(ue);
  if (not do2int(ue)) then
    error(inter, i_param, line, col);
  c:= ue.n_eval.l;
  if (c <= 0) then
    error(inter, i_param, line, col);
  write(ESC, '[255D');                  {first column -- ansi}
  write(ESC, '[', c-1, 'C');            {set c columns right -- ansi}
end;

procedure cls_ucmd;                     {CLS}
begin
  writeln(ESC, '[J');
end;

procedure random_ucmd;                  {RANDOMIZE}
begin
  randomize;
end;

procedure stop_ucmd;                    {STOP}
begin
  error(inter, i_stop, line, col);
end;

procedure end_ucmd;                     {END}
begin
  upgmend:= true;
end;

                   (***** Main Interpreter Loop *****)

procedure interpreter;
var
  uc: t_ucmd;
  pc: ^char;
begin
  ustack:= malloc(stack_size * (sizeof(t_uexpr)));
  if (ustack = NULL) then
    error(inter, i_lowmem, 0, 0);

  uprogramp:= uprogram_start;           {initialize}
  udatap:= udata_start;
  ustackp:= 0;

  upgmend:= false;
  repeat                                {main program loop}
    pc:= uprogram;
    inc(pc, uprogramp);
    uc:= p_ucmd(pc)^;
    line:= uc.line;
    col:= uc.col;
    inc(uprogramp, sizeof(t_ucmd));
    if (ord(uc.ucmd) <= max_ucmd) then
      xxx
    else
      error(inter, i_invucmd, line, col);
  until (upgmend);

  mfree(pointer(ustack));
end;

                   (********** Main **********)

procedure help;                         {print usage}
begin
  writeln('usage: cbasic -c -e -i:input_file -o:output_file');
  writeln('       cbasic --version --help');
  writeln('       -c = compile only (compile "source" into "compiled")');
  writeln('       -e = execute only (execute "compiled")');
  writeln('       else = compile and execute (compile "source" and execute it)');
  halt(0);
end;

procedure version;
begin
  writeln('extended basic: version: ', cb_version, ', ', cb_date);
  halt(0);
end;

procedure fread(p: pointer; size: long);
var
  result: int;
begin
  blockread(srcfile, p^, size, result);
  if (result <> size) then
    error(main, m_readsrc, 0, 0);
end;

procedure fwrite(p: pointer; size: long);
var
  result: int;
begin
  blockwrite(dstfile, p^, size, result);
  if (result <> size) then begin
    write(size, ':', result);
    error(main, m_writedst, 0, 0);
  end;
end;

begin
  gflags:= 0;
  assign(srcfile, '');
  assign(dstfile, '');

  for i:= 1 to paramcount do begin
    s:= paramstr(i);
    f3:= copy(s, 1, 3);
    if (s = '-c') then
      gflags:= gflags or gf_conly       {compile only}
    else if (s = '-e') then
      gflags:= gflags or gf_eonly       {execute only}
    else if (f3 = '-i:') then
      assign(srcfile, copy(s, 4, 255))  {input file}
    else if (f3 = '-o:') then
      assign(dstfile, copy(s, 4, 255))  {output file}
    else if (s = '--version') then
      version                           {version}
    else if ((s = '-?') or (s = '--help')) then
      help                              {help}
    else
      error(main, m_param, 0, 0);
  end;

  {$I-}
  reset(srcfile, 1);
  if (ioresult <> 0) then
    error(main, m_opensrc, 0, 0);
  rewrite(dstfile, 1);
  if (ioresult <> 0) then
    error(main, m_opendst, 0, 0);
  {$I+}

  if ((gflags and gf_eonly) = 0) then begin
    compiler;                           {compiler}
    writeln;
    write(uprogram_top, ' + ', udata_top, ' bytes (', line-1, ' line');
    if (line-1 > 1) then
      write('s');
    writeln(') successfully compiled.');
  end;

  if ((gflags and gf_conly) <> 0) then begin
    f3:= cb_ver_fsign;
    fwrite(@f3[1], 3);
    fwrite(@uprogram_start, sizeof(uprogram_start));
    fwrite(@udata_start, sizeof(udata_start));
    fwrite(@uprogram_top, sizeof(uprogram_top));
    fwrite(uprogram, (uprogram_top));
    fwrite(@uvars_top, sizeof(uvars_top));
    fwrite(uvars, (uvars_top));
  end;

  if ((gflags and gf_eonly) <> 0) then begin
    f3:= 'XXX';                         {size = 3}
    fread(@f3[1], 3);
    if (f3 <> cb_ver_fsign) then
      error(main, m_wfile, 0, 0);
    fread(@uprogram_start, sizeof(uprogram_start));
    fread(@udata_start, sizeof(udata_start));
    fread(@uprogram_top, sizeof(uprogram_top));
    uprogram:= malloc(uprogram_top);
    if (uprogram = NULL) then
      error(main, m_lowmem, line, col);
    fread(uprogram, uprogram_top);
    fread(@uvars_top, sizeof(uvars_top));
    uvars:= malloc(uvars_top);
    if (uvars = NULL) then
      error(main, m_lowmem, line, col);
    fread(uvars, uvars_top);
  end;

  if ((gflags and gf_conly) = 0) then
    interpreter;                        {interpreter}

  halt(0);
end.
